TAG_CRM_CONFIG = "crm_config"
TAG_RESOURCE_PRIMITIVE = "primitive"
TAG_RESOURCE_GROUP = "group"
TAG_RESOURCE_BUNDLE = "bundle"
TAG_RESOURCE_CLONE = "clone"
TAG_RESOURCE_MASTER = "master"
TAG_TAG = "tag"
TAG_OBJREF = "obj_ref"
