from typing import NewType

FileTypeCode = NewType("FileTypeCode", str)

BOOTH_CONFIG = FileTypeCode("BOOTH_CONFIG")
BOOTH_KEY = FileTypeCode("BOOTH_KEY")
CIB = FileTypeCode("CIB")
COROSYNC_AUTHKEY = FileTypeCode("COROSYNC_AUTHKEY")
COROSYNC_CONF = FileTypeCode("COROSYNC_CONF")
COROSYNC_QDEVICE_NSSDB = FileTypeCode("COROSYNC_QDEVICE_NSSDB")
COROSYNC_QNETD_NSSDB = FileTypeCode("COROSYNC_QNETD_NSSDB")
PACEMAKER_AUTHKEY = FileTypeCode("PACEMAKER_AUTHKEY")
PCSD_ENVIRONMENT_CONFIG = FileTypeCode("PCSD_ENVIRONMENT_CONFIG")
PCSD_SSL_CERT = FileTypeCode("PCSD_SSL_CERT")
PCSD_SSL_KEY = FileTypeCode("PCSD_SSL_KEY")
PCS_KNOWN_HOSTS = FileTypeCode("PCS_KNOWN_HOSTS")
PCS_SETTINGS_CONF = FileTypeCode("PCS_SETTINGS_CONF")
PCS_DR_CONFIG = FileTypeCode("PCS_DR_CONFIG")
PCS_USERS_CONF = FileTypeCode("PCS_USERS_CONF")
