from unittest import TestCase


class SendAllConfigToNodeTest(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.cluster.test_add_nodes.AddNodeFull
        pcs_test.tier0.lib.commands.cluster.test_add_nodes
            .FailureBoothConfigsDistribution
    """
